﻿using System;
using System.Windows.Controls;
using System.Windows.Media;
using System.Windows.Shapes;
using System.Windows.Threading;
using Telerik.Windows.Controls;
using Telerik.Windows.Controls.Map;

namespace DayNightTerminator
{
	public partial class MainPage : UserControl
	{
		private const double degreesToRadians = Math.PI / 180.0;
		private const double radiansToDegrees = 180.0 / Math.PI;

		private DateTime time;

		private ISpatialReference spatialReference;

		private DispatcherTimer timer;

		private PolylineData polyline;
		private Ellipse sun;

		public MainPage()
		{
			this.spatialReference = new MercatorProjection();
			this.time = DateTime.Now;

			InitializeComponent();

			this.GenerateGrid();
			this.CreateTerminator();
			this.RefreshTerminator();

			this.timer = new DispatcherTimer();

			this.timer.Interval = TimeSpan.FromMinutes(1);
			// this.timer.Interval = TimeSpan.FromSeconds(1);

			this.timer.Tick += timer_Tick;
			this.timer.Start();
		}

		private void timer_Tick(object sender, EventArgs e)
		{
			this.timer.Stop();

			this.time = DateTime.Now;
			// this.time = this.time.AddHours(1);

			this.RefreshTerminator();
			this.timer.Start();
		}

		private void GenerateGrid()
		{
			for (double longitude = -180 + 15; longitude < 180; longitude += 15)
			{
				var line = new LineData()
				{
					Point1 = new Location(this.spatialReference.MinLatitude, longitude),
					Point2 = new Location(this.spatialReference.MaxLatitude, longitude),
					ShapeFill = new MapShapeFill()
					{
						StrokeThickness = 0.5,
						StrokeDashArray = new DoubleCollection() { 2, 2 },
						Stroke = new SolidColorBrush(Color.FromArgb(0x5f, 0x1f, 0x1f, 0x1f))
					}
				};
				this.gridLayer.Items.Add(line);
			}

			for (double latitude = -60; latitude <= 60; latitude += 30)
			{
				var line = new LineData()
				{
					Point1 = new Location(latitude, this.spatialReference.MinLongitude),
					Point2 = new Location(latitude, this.spatialReference.MaxLongitude),
					ShapeFill = new MapShapeFill()
					{
						StrokeThickness = 0.5,
						StrokeDashArray = new DoubleCollection() { 2, 2 },
						Stroke = new SolidColorBrush(Color.FromArgb(0x5f, 0x1f, 0x1f, 0x1f))
					}
				};
				this.gridLayer.Items.Add(line);
			}
		}

		private void CreateTerminator()
		{
			this.polyline = new PolylineData()
			{
				ShapeFill = new MapShapeFill()
				{
					StrokeThickness = 1,
					Stroke = new SolidColorBrush(Color.FromArgb(0x5f, 0x1f, 0x1f, 0x1f)),
					Fill = new SolidColorBrush(Color.FromArgb(0x7f, 0x7f, 0x7f, 0x7f))
				}
			};

			this.terminatorLayer.Items.Add(this.polyline);

			this.sun = new Ellipse()
			{
				Width = 15,
				Height = 15,
				StrokeThickness = 2,
				Stroke = new SolidColorBrush(Color.FromArgb(0xff, 0xcf, 0xbf, 0)),
				Fill = new SolidColorBrush(Color.FromArgb(0xaf, 0xff, 0xff, 0))
			};
			MapLayer.SetHotSpot(this.sun, new HotSpot()
			{
				X = 0.5,
				Y = 0.5
			});

			this.terminatorLayer.Items.Add(this.sun);
		}

		private void RefreshTerminator()
		{
			Location sunLocation = this.CalculateSunCoordinates();
			MapLayer.SetLocation(this.sun, sunLocation);

			var points = new LocationCollection();
			bool southOrientation = sunLocation.Latitude > 0;
			this.AddStartPoint(points, southOrientation);

			for (double longitude = -180; longitude <= 180; longitude += 1)
			{
				this.AddPoint(points, longitude, sunLocation);
			}

			this.AddEndPoints(points, southOrientation);

			this.polyline.Points = points;
		}

		private void AddPoint(LocationCollection points, double longitude, Location sunLocation)
		{
			Location location = new Location();
			location.Longitude = longitude - sunLocation.Longitude;
			if (location.Longitude > 180)
			{
				location.Longitude = location.Longitude - 360;
			}
			else if (location.Longitude < -180)
			{
				location.Longitude = location.Longitude + 360;
			}

			location.Latitude = radiansToDegrees * Math.Atan(-Math.Cos(location.Longitude * degreesToRadians) / Math.Tan(sunLocation.Latitude * degreesToRadians));
			if (location.Latitude > this.spatialReference.MaxLatitude)
			{
				location.Latitude = this.spatialReference.MaxLatitude;
			}
			else if (location.Latitude < this.spatialReference.MinLatitude)
			{
				location.Latitude = this.spatialReference.MinLatitude;
			}

			location.Longitude = longitude;
			points.Add(location);
		}

		private void AddEndPoints(LocationCollection points, bool southOrientation)
		{
			if (southOrientation)
			{
				points.Add(new Location(this.spatialReference.MinLatitude, 180));
				points.Add(new Location(this.spatialReference.MinLatitude, -180));
			}
			else
			{
				points.Add(new Location(this.spatialReference.MaxLatitude, 180));
				points.Add(new Location(this.spatialReference.MaxLatitude, -180));
			}
		}

		private void AddStartPoint(LocationCollection points, bool southOrientation)
		{
			if (southOrientation)
			{
				points.Add(new Location(this.spatialReference.MinLatitude, -180));
			}
			else
			{
				points.Add(new Location(this.spatialReference.MaxLatitude, -180));
			}
		}

		private Location CalculateSunCoordinates()
		{
			DateTime universalTime = time.ToUniversalTime();
			double minute = universalTime.Minute;
			double hour = universalTime.Hour;

			double uT = hour + minute / 60d;
			double t = GetTime(universalTime, uT);
			double m = 357.528 + (35999.05 * t) - (0.0001559 * t * t) - (0.00000048 * t * t * t);
			m = AngleTo360(m) * degreesToRadians;

			double c = (1.915 * Math.Sin(m)) + (0.02 * Math.Sin(m * 2d));
			double l = 280.46 + (36000.77 * t) + c;
			l = AngleTo360(l);

			double alpha = l - 2.466 * Math.Sin(l * degreesToRadians * 2d) + 0.053 * Math.Sin(l * degreesToRadians * 4d);
			double obliquity = 23.4393 - 0.013 * t;
			double declination = Math.Atan(Math.Tan(obliquity * degreesToRadians) * Math.Sin(alpha * degreesToRadians)) * radiansToDegrees;

			double hourAngle = (uT * 15d) - 180d - c + l - alpha;
			hourAngle = AngleTo180(hourAngle);

			Location location = new Location(declination, -hourAngle);

			return location;
		}

		private double GetTime(DateTime universalTime, double uT)
		{
			double day = universalTime.Day;
			double month = universalTime.Month;
			double year = universalTime.Year;
			double julianYear, julianMonth;
			if (month > 2)
			{
				julianYear = year;
				julianMonth = month - 3;
			}
			else
			{
				julianYear = year - 1;
				julianMonth = month + 9;
			}

			double t = (day + Math.Floor(30.6 * julianMonth + 0.5) + Math.Floor(365.25 * (julianYear - 1976)) - 8707.5 + (uT / 24d)) / 36525d;
			return t;
		}

		private double AngleTo360(double value)
		{
			value = value - Math.Floor(value / 360d) * 360d;
			return value;
		}

		private double AngleTo180(double value)
		{
			value = AngleTo360(value);
			if (value > 180)
			{
				value = value - 360;
			}
			return value;
		}
	}
}
